import json
import os

from ptera import ABSENT

format_registry = {}


def register_extension(*extensions):
    def deco(fn):
        for ext in extensions:
            format_registry[ext] = fn
        return fn

    return deco


@register_extension("json")
def formatter_json():
    return {
        "load": json.load,
        "dump": json.dump,
    }


@register_extension("yaml")
def formatter_yaml():
    try:
        import yaml
    except ModuleNotFoundError:  # pragma: no cover
        raise OSError(
            "Please install the 'pyyaml' module to read '.yaml' files"
        )
    return {
        "load": yaml.full_load,
        "dump": yaml.dump,
    }


@register_extension("toml")
def formatter_toml():
    try:
        import toml
    except ModuleNotFoundError:  # pragma: no cover
        raise OSError("Please install the 'toml' module to read '.toml' files")
    return {
        "load": toml.load,
        "dump": toml.dump,
    }


def _read_cfg(file):
    import configparser

    parser = configparser.ConfigParser()
    s = file.read()
    parser.read_string(s)
    if parser.sections() != ["default"]:
        raise OSError(
            "A cfg/ini file containing options must only contain"
            " the [default] section"
        )
    return dict(parser["default"])


@register_extension("cfg", "ini")
def formatter_cfg():
    return {
        "load": _read_cfg,
    }


class ConfigFile:
    def __init__(self, filename, registry=format_registry):
        self.filename = os.path.expanduser(filename)
        ext = filename.split(".")[-1]
        fmt = registry.get(ext, None)
        if fmt is None:
            raise OSError(f"Cannot read file format: '{ext}'")
        self.format = fmt()

    def read(self, default=ABSENT):
        try:
            with open(self.filename) as f:
                return self.format["load"](f)
        except FileNotFoundError:
            if default is ABSENT:
                raise
            return default

    def write(self, contents):
        os.makedirs(os.path.dirname(self.filename), exist_ok=True)
        with open(self.filename, "w") as f:
            return self.format["dump"](contents, f)


def config(filename):
    if any(filename.startswith(c) for c in '{["'):
        return json.loads(filename)
    else:
        return ConfigFile(filename).read()



import argparse
import os
import re
import sys
from collections import defaultdict
from contextlib import contextmanager
from itertools import count
from types import FunctionType, SimpleNamespace

from ptera import (
    ABSENT,
    BaseOverlay,
    PteraFunction,
    Tag,
    TagSet,
    match_tag,
    select,
    tag,
    tooled,
)
from ptera.selfless import PreState, PteraNameError

from .config import ConfigFile

Argument = tag.Argument


_count = count()


def _catalogue(seen, results, fn):
    if id(fn) in seen:
        return

    seen.add(id(fn))

    if isinstance(fn, PteraFunction):
        state = fn.state.state if isinstance(fn.state, PreState) else fn.state
        tst = type(state)
        res = tst.__info__
        results[fn] = res
        for name, ann in res.items():
            val = getattr(state, name, ABSENT)
            _catalogue(seen, results, val)

    elif isinstance(fn, (list, tuple)):
        for entry in fn:
            _catalogue(seen, results, entry)


def catalogue(functions):
    results = {}
    _catalogue(set(), results, functions)
    return results


def _find_configurable(catalogue, tag):
    rval = defaultdict(dict)
    for fn, variables in catalogue.items():
        for name, data in variables.items():
            ann = data["annotation"]
            if match_tag(tag, ann):
                rval[name][fn] = data
    return rval


class ArgsExpander:
    def __init__(self, prefix, default_file=None):
        self.prefix = prefix
        self.default_file = default_file
        if default_file:
            assert self.prefix

    def _generate_args_from_dict(self, contents):
        results = []
        for key, value in contents.items():
            if key == "#include":
                if not isinstance(value, (list, tuple)):
                    value = [value]
                for other_filename in value:
                    results.extend(
                        self._generate_args_from_file(other_filename)
                    )

            elif key == "#command":
                results.insert(0, value)

            elif key.startswith("-"):
                results.append(key)
                results.append(str(value))

            elif isinstance(value, bool):
                if value:
                    results.append(f"--{key}")
                else:
                    results.append(f"--no-{key}")

            else:
                results.append(f"--{key}")
                results.append(str(value))
        return results

    def _generate_args_from_file(self, filename):
        contents = ConfigFile(filename).read()
        return self._generate_args_from_dict(contents)

    def expand(self, argv):
        if self.default_file:
            if os.path.exists(self.default_file):
                pfx = self.prefix[0]
                argv.insert(0, f"{pfx}{self.default_file}")

        new_args = []
        for arg in argv:
            if isinstance(arg, dict):
                new_args.extend(self._generate_args_from_dict(arg))
            elif not arg or arg[0] not in self.prefix:
                new_args.append(arg)
            else:
                new_args.extend(self._generate_args_from_file(arg[1:]))
        return new_args

    def __call__(self, argv, *, parser):
        try:
            return self.expand(argv)
        except OSError:
            err = sys.exc_info()[1]
            parser.error(str(err))


def parse_options(parser, *, argv=None, expand=None):
    if isinstance(argv, argparse.Namespace):
        args = argv
    elif isinstance(argv, dict):
        args = argparse.Namespace(**argv)
    else:
        argv = sys.argv[1:] if argv is None else argv
        if expand:
            argv = expand(argv, parser=parser)
        args = parser.parse_args(argv)
    return args


class Configurator:
    def __init__(
        self,
        *,
        argparser,
        entry_point=None,
        tag=Argument,
        description=None,
        eval_env=None,
        expand=None,
    ):
        cg = catalogue(entry_point)
        self.tag = tag
        self.names = _find_configurable(cg, tag)
        self.argparser = argparser
        self._fill_argparser()
        self.eval_env = eval_env
        self.expand = expand

    def _analyze_entry(self, name, data):
        docs = set()
        for fn, entry in data.items():
            if entry["doc"]:
                docs.add(entry["doc"])
            else:
                docs.add(f"Parameter in {fn}")

        loc = None
        if len(data) == 1:
            (sole_data,) = list(data.values())
            loc = sole_data["location"]

        optname = name.replace("_", "-")
        typ = []
        for x in data.values():
            ann = x["annotation"]
            if isinstance(ann, TagSet):
                members = ann.members
            else:
                members = [ann]
            for m in members:
                if not isinstance(m, Tag):
                    typ.append(m)
        if len(typ) != 1:
            typ = None
        else:
            (typ,) = typ

        default_opt = f"--{optname}"
        aliases = [default_opt]
        nargs = False
        positional = False
        group = None
        negate = None
        metavar = None
        optdoc = []
        for entry in docs:
            new_entry = []
            for line in entry.split("\n"):
                m = re.match(r"\[([A-Za-z0-9_-]+)(:.*)?\]", line)
                if m:
                    command, arg = m.groups()
                    command = command.lower()
                    arg = arg and arg[1:].strip()
                    if command in ["alias", "aliases"]:
                        aliases.extend(re.split(r"[ ,;]+", arg))
                    elif command in ["option", "options"]:
                        aliases = re.split(r"[ ,;]+", arg)
                    elif command == "group":
                        group = arg
                    elif command == "negate":
                        negate = arg or True
                    elif command == "metavar":
                        metavar = arg
                    elif command == "remainder":
                        positional = True
                        nargs = argparse.REMAINDER
                    elif command in ["nargs", "positional"]:
                        positional = command == "positional"
                        nargs = arg or None
                        try:
                            nargs = int(nargs)
                        except Exception:
                            pass
                else:
                    new_entry.append(line)
            optdoc.append("\n".join(new_entry))

        opts = SimpleNamespace(
            positional=positional,
            metavar=metavar,
            negate=negate,
            group=group,
            name=name,
            optname=optname,
            doc=optdoc,
            nargs=nargs,
            type=typ,
            aliases=aliases,
            loc=loc,
            has_default_opt_name=(aliases[0] == default_opt),
        )
        for x in data.values():
            x["coleo_options"] = opts
        return opts

    def _fill_argparser(self):
        groups = {}

        entries = [
            self._analyze_entry(name, data) for name, data in self.names.items()
        ]

        positional = list(
            sorted(
                (entry for entry in entries if entry.positional),
                key=lambda entry: entry.loc[-1] if entry.loc else -1,
            )
        )
        if len(positional) > 1:
            if any(entry.loc is None for entry in positional):
                raise Exception(
                    "Positional arguments cannot be defined in multiple"
                    " functions."
                )
            loc0 = positional[0].loc[:2]
            if not all(
                entry.loc is not None and entry.loc[:2] == loc0
                for entry in positional
            ):
                raise Exception(
                    "All positional arguments must be defined in the same"
                    " function."
                )

        nonpositional = list(
            sorted(
                (entry for entry in entries if not entry.positional),
                key=lambda entry: entry.name,
            )
        )

        entries = positional + nonpositional

        for entry in entries:
            name = entry.name
            typ = entry.type
            aliases = entry.aliases
            nargs = entry.nargs
            optdoc = entry.doc
            optname = entry.optname

            if entry.group:
                if entry.group not in groups:
                    groups[entry.group] = self.argparser.add_argument_group(
                        title=entry.group
                    )
                group = groups[entry.group]
            else:
                group = self.argparser

            if typ is bool:
                if entry.negate is not None:
                    group = self.argparser.add_mutually_exclusive_group()

                group.add_argument(
                    *aliases,
                    dest=name,
                    action="store_true",
                    help="; ".join(optdoc),
                )

                if entry.negate is not None:
                    group.add_argument(
                        f"--no-{optname}",
                        dest=name,
                        action="store_false",
                        help=f"Set --{optname} to False"
                        if entry.negate is True
                        else entry.negate,
                    )
            else:
                if entry.positional:
                    group.add_argument(
                        name,
                        type=self.resolver(typ or None),
                        action="store",
                        nargs=nargs,
                        metavar=name.upper(),
                        default=[] if nargs == "*" else None,
                        help="; ".join(optdoc),
                    )
                else:
                    _metavars = {
                        int: "NUM",
                        float: "NUM",
                        argparse.FileType: "FILE",
                    }
                    ttyp = typ if isinstance(typ, type) else type(typ)
                    mv = entry.metavar or _metavars.get(ttyp, "VALUE")
                    nargs_kw = {"nargs": nargs} if nargs else {}
                    group.add_argument(
                        *aliases,
                        dest=name,
                        type=self.resolver(typ or None),
                        action="store",
                        metavar=mv,
                        help="; ".join(optdoc),
                        **nargs_kw,
                    )

    def resolver(self, typ):
        def resolve(arg):
            if typ is str:
                return arg
            elif re.match(r"^:[A-Za-z_0-9]+:", arg):
                _, modname, code = arg.split(":", 2)
                mod = __import__(modname)
                return eval(code, vars(mod))
            elif arg.startswith(":"):
                if not self.eval_env:
                    raise Exception(f"No environment to evaluate {arg}")
                return eval(arg[1:], self.eval_env)
            else:
                return arg if typ is None else typ(arg)

        resolve.__name__ = getattr(typ, "__name__", str(typ))
        return resolve

    @contextmanager
    def __call__(self, argv=None):
        opts = parse_options(self.argparser, argv=argv, expand=self.expand)
        opts = {k: v for k, v in vars(opts).items() if not k.startswith("#")}
        with _setvars(opts, self.tag):
            yield opts


def _getdoc(obj):
    if isinstance(obj, dict):
        return obj.get("__doc__", None)
    else:
        return getattr(obj, "__doc__", None)


def _auto_cli_helper(parser, entry, **kwargs):
    if isinstance(entry, dict):
        subparsers = parser.add_subparsers()
        for name, subentry in entry.items():
            if name == "__doc__":
                continue
            subparser = subparsers.add_parser(
                name, help=_getdoc(subentry), argument_default=argparse.SUPPRESS
            )
            _auto_cli_helper(subparser, subentry)
    else:
        if isinstance(entry, FunctionType):
            entry = tooled(entry)
        if not isinstance(entry, PteraFunction):
            raise TypeError(f"Expected a dict or a function, not {type(entry)}")
        cfg = Configurator(entry_point=entry, argparser=parser, **kwargs)
        parser.set_defaults(**{"#cfg": (cfg, entry)})


def _setvars(values, tag):
    def _resolver(value):
        return lambda **_: value

    return BaseOverlay(
        {
            select(f"{name}:##X", env={"##X": tag}): {"value": _resolver(value)}
            for name, value in values.items()
        }
    )


def setvars(**values):
    return _setvars(values, tag=Argument)


def auto_cli(
    entry,
    args=(),
    *,
    argv=None,
    entry_point=None,
    tag=Argument,
    description=None,
    eval_env=None,
    expand=None,
    print_result=True,
    return_split=False,
):
    if expand is None or isinstance(expand, str):
        expand = ArgsExpander(prefix=expand or "", default_file=None,)

    parser = argparse.ArgumentParser(
        description=description or _getdoc(entry),
        argument_default=argparse.SUPPRESS,
    )
    _auto_cli_helper(parser, entry, tag=tag, eval_env=eval_env)
    opts = parse_options(parser, argv=argv, expand=expand)
    cfg, fn = getattr(opts, "#cfg", (None, None))
    if cfg is None:
        parser.print_help()
        sys.exit(1)

    def thunk(opts=opts, args=args):
        with cfg(opts):
            try:
                result = fn(*args)
            except PteraNameError as err:
                optinfo = err.info().get("coleo_options", None)
                if optinfo:
                    optname = optinfo.aliases[0]
                    print(
                        f"error: missing value for required argument: {optname}",
                        file=sys.stderr,
                    )
                    sys.exit(1)
                else:
                    raise
            if print_result and result is not None:
                print(result)
            return result

    if return_split:
        return opts, thunk
    else:
        return thunk(opts)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['coleo']

package_data = \
{'': ['*']}

install_requires = \
['ptera>=0.2.4,<0.3.0']

setup_kwargs = {
    'name': 'coleo',
    'version': '0.1.6',
    'description': 'The nicest way to develop a command-line interface',
    'long_description': '\n# Coleo\n\nColeo is a minimum-effort way to create a command-line interface in Python.\n\n* Declare options where they are used.\n* Scale easily to extensive CLIs with dozens of subcommands and options.\n\n\n## Basic usage\n\nFirst, define a command line interface as follows:\n\n```python\nfrom coleo import Argument, auto_cli, default\n\ndef main():\n    # The greeting\n    greeting: Argument = default("Hello")\n\n    # The name to greet\n    name: Argument = default("you")\n\n    return f"{greeting}, {name}!"\n\nif __name__ == "__main__":\n    auto_cli(main)\n```\n\nThen you may run it like this on the command line:\n\n```bash\n$ python hello.py\nHello, you!\n$ python hello.py --name Luke\nHello, Luke!\n$ python hello.py --name Luke --greeting "Happy birthday"\nHappy birthday, Luke!\n$ python hello.py -h\nusage: hello.py [-h] [--greeting VALUE] [--name VALUE]\n\noptional arguments:\n  -h, --help        show this help message and exit\n  --greeting VALUE  The greeting\n  --name VALUE      The name to greet\n```\n\n* Any variable annotated with `Argument` will become an option.\n* You can provide a default value with `default(value)`, although you don\'t have to, if the argument is required.\n* If there is a comment above the variable, it will be used as documentation for the option.\n\n\n## Argument types\n\nBy default, all arguments are interpreted as strings, but you can easily give a different type to an argument:\n\n```python\ndef main():\n    # This argument will be converted to an int\n    x: Argument & int\n    # This argument will be converted to a float\n    y: Argument & float\n    return x + y\n```\n\n**Boolean flags**\n\nIf the type is bool, the option will take no argument, and the `--no-<optname>` option is added to negate the value. For example:\n\n```python\ndef main():\n    flag: Argument & bool\n    return "yes!" if flag else "no!"\n```\n\nUse it like this:\n\n```bash\n$ python script.py --flag\nyes!\n$ python script.py --no-flag\nno!\n```\n\n**Files**\n\nUse `coleo.FileType` (or `argparse.FileType`, it\'s the same thing) to open a file to read from or to write to:\n\n```python\ndef main():\n    grocery_list: Argument & coleo.FileType("r")\n    with grocery_list as f:\n        for food in f.readlines():\n            print(f"Gotta buy some {food}")\n```\n\n**Config**\n\nYou can manipulate configuration files with `coleo.config` or `coleo.ConfigFile`:\n\n```python\ndef main():\n    # ConfigFile lets you read or write a configuration file\n    book: Argument & ConfigFile\n    contents = book.read()\n    contents["xyz"] = "abc"\n    book.write(contents)\n\n    # config will read the file for you or parse the argument as JSON\n    magazine: Argument & config\n    print(magazine)\n```\n\nUse it simply like this:\n\n```bash\n$ python librarian.py --book alice.json --magazine vogue.json\n$ python librarian.py --book history.yaml --magazine gamez.toml\n$ python librarian.py --book physics.json --magazine \'{"a": 1, "b": 2}\'\n# etc\n```\n\nSupported extensions are `json`, `yaml` and `toml` (the latter two require installing the `pyyaml` or `toml` packages).\n\n\n**Other**\n\nAny function can be used as a "type" for an argument. So for example, if you want to be able to provide lists and dictionaries on the command line you can simply use `json.loads` (although `coleo.config` is usually better, because it can also read files, in various formats):\n\n```python\ndef main():\n    obj: Argument & json.loads\n    return type(obj).__name__\n```\n\n```bash\n$ python json.py --obj 1\nint\n$ python json.py --obj \'"hello"\'\nstr\n$ python json.py --obj \'{"a": 1, "b": 2}\'\ndict\n```\n\nIf you\'re feeling super feisty and care nothing about safety, you can even use `eval`:\n\n```python\ndef main():\n    obj: Argument & eval\n    return type(obj).__name__\n```\n\n```bash\n$ python eval.py --obj "1 + 2"\nint\n$ python eval.py --obj "lambda x: x + 1"\nfunction\n```\n\n\n## Customization\n\nUsing comments of the form `# [<instruction>: <args ...>]` you can customize the option parser:\n\n```python\ndef main():\n    # This argument can be given as either --greeting or -g\n    # [alias: -g]\n    greeting: Argument = default("Hello")\n\n    # This argument is positional\n    # [positional]\n    name: Argument = default("you")\n\n    # This argument can only be given as -n\n    # [options: -n]\n    ntimes: Argument & int = default(1)\n\n    for i in range(ntimes):\n        print(f"{greeting}, {name}!")\n```\n\nThe above would be used like this:\n\n```bash\n$ python hello.py Alice -g Greetings -n 2\nGreetings, Alice!\nGreetings, Alice!\n```\n\nThe following customizations are available:\n\n* `[alias: ...]` defines one or several options that are aliases for the main one. Options are separated by spaces, commas or semicolons.\n* `[options: ...]` defines one or several options for this argument, which *override* the default one. Options are separated by spaces, commas or semicolons.\n* `[positional]` defines one positional argument.\n  * `[positional: n]`: n positional arguments (a list is returned).\n  * `[positional: ?]`: one optional positional argument\n  * `[positional: *]`: zero or more positional arguments\n  * `[positional: +]`: one or more positional arguments\n\nIt is currently not possible to define multiple arguments as positional, although you can simulate it by using e.g. `[positional: 2]` on a single argument to get a pair, and then you can get the elements of the list.\n\n\n## Subcommands\n\nYou can create an interface with a hierarchy of subcommands by passing a dictionary to `auto_cli`:\n\n```python\ndef add():\n    x: Argument & int\n    y: Argument & int\n    return x + y\n\ndef mul():\n    x: Argument & int\n    y: Argument & int\n    return x * y\n\ndef pow():\n    base: Argument & int\n    exponent: Argument & int\n    return base ** exponent\n\ndef greet():\n    greeting: Argument = default("Hello")\n    name: Argument = default("you")\n    return f"{greeting}, {name}!"\n\nif __name__ == "__main__":\n    auto_cli({\n        "calc": {\n            "__doc__": "Calculate something!",\n            "add": add,\n            "mul": mul,\n            "pow": pow,\n        },\n        "greet": greet,\n    })\n```\n\nThen you may use it like this:\n\n```bash\n$ python multi.py greet --name Alice --greeting Hi\nHi, Alice!\n$ python multi.py calc add --x=3 --y=8\n11\n```\n\n\n## Sharing arguments\n\nIt is possible to share behavior and arguments between subcommands, or to split complex functionality into multiple pieces. For example, maybe multiple subcommands in your application require an API key, which can either be given on the command line or can be read from a file. This is how you would share this behavior across all subcommands:\n\n```python\nfrom coleo import Argument, auto_cli, default, tooled\n\n@tooled\ndef apikey():\n    # The API key to use\n    key: Argument = default(None)\n    if key is None:\n        # If no key parameter is given on the command line, try to read it from\n        # some standard location.\n        key = config("~/.config/myapp/config.json")["key"]\n    return key\n\ndef search():\n    interface = Application(apikey())\n    query: Argument\n    return interface.search(query)\n\ndef install():\n    interface = Application(apikey())\n    package: Argument\n    return interface.install(package)\n\nif __name__ == "__main__":\n    auto_cli({"search": search, "install": install})\n```\n\nIf a function is decorated with `@tooled` and is called from one of the main functions (or from another tooled function), Coleo will search for arguments in that function too. Thus any subcommand that calls `apikey()` will gain a `--key` option.\n\nIn addition to this, you can "share" arguments by defining the same argument with the same type in multiple functions. Coleo will set all of them to the same value.\n\nFor example, in the example above you could easily let the user specify the path to the file that contains the key, simply by replacing\n\n```python\nkey = config("~/.config/myapp/config.json")["key"]\n\n# ==>\n\nconfig_path: Argument = default("~/.config/myapp/config.json")\nkey = config(config_path)["key"]\n```\n\nAnd that `config_path` argument could, of course, be declared in any other function that needs to read some configuration value.\n\n\n## Non-CLI usage\n\nIt is possible to set arguments without `auto_cli` with `setvars`:\n\n```python\nfrom coleo import Argument, setvars, tooled\n\n@tooled\ndef greet():\n    greeting: Argument = default("Hello")\n    name: Argument = default("you")\n    return f"{greeting} {name}!"\n\nwith setvars(greeting="Hi", name="Bob"):\n    assert greet() == "Hi bob!"\n```\n\nNote:\n\n* With `setvars`, you *must* decorate the function with `@tooled` (this is something `auto_cli` does on your behalf).\n* `setvars` entirely bypasses the option parsing and the type annotations will not be used to wrap these values. In other words, if a variable is annotated `Argument & int` and you provide the value "1", it will remain a string.\n\n\n### Using with Ptera\n\nColeo is based on [Ptera](https://github.com/mila-iqia/ptera) and all of Ptera\'s functionality is de facto available on functions marked as `@tooled`. For example, using the example above:\n\n```python\n# Set the variables in the greet function -- it\'s a bit like making an object\nhibob = greet.new(greeting="Hi", name="Bob")\nassert hibob() == "Hi Bob!"\n\n# Same as above but this would also change greeting/name in any other function\n# that is called by greet, and so on recursively (a bit like dynamic scoping)\nhibob = greet.tweaking({"greeting": "Hi", "name": "Bob"})\nassert hibob() == "Hi Bob!"\n\n# More complex behavior\nfrom ptera import overlay\nwith overlay.tweaking({\n    "greet(greeting=\'Bonjour\') > name": "Toto"\n}):\n    assert greet() == "Hello you!"\n    assert greet.new(greeting="Hi")() == "Hi you!"\n    assert greet.new(greeting="Bonjour")() == "Bonjour toto!"\n```\n\nRead the documentation for [Ptera](https://github.com/mila-iqia/ptera) for more information. Note that Ptera is not limited to variables tagged `Argument`, it can manipulate *any* variable in a tooled function.\n',
    'author': 'Olivier Breuleux',
    'author_email': 'breuleux@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/breuleux/coleo',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
